/* ------------------------------------------------------------------------
 * $Id: client.cc,v 1.6 2001/12/07 13:09:33 elm Exp $
 *
 * This file is part of 3Dwm: The Three-Dimensional User Environment.
 *
 * 3Dwm: The Three-Dimensional User Environment:
 *	<http://www.3dwm.org>
 *
 * Chalmers Medialab
 * 	<http://www.medialab.chalmers.se>
 * 
 * ------------------------------------------------------------------------
 * File created 2001-07-31 by Niklas Elmqvist.
 *
 * Copyright (c) 2001 Niklas Elmqvist <elm@3dwm.org>.
 * ------------------------------------------------------------------------
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * ------------------------------------------------------------------------
 */

// -- System Includes
#include <time.h>
#include <iostream>
#include <unistd.h>

// -- 3Dwm Includes
#include <Nobel++/Nobel.hh>
#include <Nobel/Solid.hh>
#include <Nobel/SolidKit.hh>
#include <Nobel/SolidContainer.hh>
#include <Nobel/Appearance.hh>
#include <Nobel/GeometryKit.hh>
#include <Nobel/PrimitiveKit.hh>

#include <Celsius/GetOpt.hh>

using namespace Nobel;

// -- Defines
#define VERSION		"0.2"

// -- Class Declarations

class Clock {
public:
    
    /**
     * Constructor.
     **/
    Clock(int &argc, char **argv, const GetOpt &opt)
	: _client("Clock", argc, argv),
	  _opt(opt) {
	
	// Resolve the necessary kits
	_geo_kit = _client.resolve<GeometryKit>(Nobelxx::name<GeometryKit>());
	_solid_kit = _client.resolve<SolidKit>(Nobelxx::name<SolidKit>());
	_prim_kit = _client.resolve<PrimitiveKit>
	    (Nobelxx::name<PrimitiveKit>());
	
	// Build the clock geometry
	buildClock(Node_var(_client.root()));
    }
    
    /**
     * Destructor.
     **/
    ~Clock() {
	// empty
    }
    
    void run() {

	while (true) {
	    
	    // Retrieve a time value
	    time_t abstime = time(0);
	    tm *t = localtime(&abstime);

	    // Update the clock hand positions
	    setHands(t->tm_hour, t->tm_min, t->tm_sec);
	    
	    // Sleep for a while
	    sleep(5);
	}
    }

private:

    bool isnumerical(char c) {
	return isdigit(c) || c == '.' || c == '-';
    }
    
    std::string parseValue(const std::string &str, float &value, char delim) {
	std::string::size_type begin = 0;
	while (begin < str.length() && str[begin] == delim)
	    begin++;
	std::string::size_type end = begin;
	while (end < str.length() && isnumerical(str[end]))
	    end++;
	if (begin == end) return str.substr(begin);
	std::string number = str.substr(begin, end);
	value = strtod(number.c_str(), NULL);
	return str.substr(end);
    }

    /**
     * Application geometry. 
     **/ 
    struct Geometry {
	float width, height, depth;
	float x, y, z;
	float xrot, yrot, zrot;
	Geometry(float _width = 0, float _height = 0, float _depth = 0,
		 float _x = 0, float _y = 0, float _z = 0,
		 float _xrot = 0, float _yrot = 0, float _zrot = 0) 
	    : x(_x), y(_y), z(_z),
	      width(_width), height(_height), depth(_depth),
	      xrot(_xrot), yrot(_yrot), zrot(_zrot)
	{ }
    };
    
    void parseGeometry(const std::string &s, Geometry &geo) {
	std::string str = s;

	// Parse position
	str = parseValue(str, geo.width, 'x');
	str = parseValue(str, geo.height, 'x');
	str = parseValue(str, geo.depth, 'x');
	
	// Parse dimensions
	str = parseValue(str, geo.x, '+');
	str = parseValue(str, geo.y, '+');
	str = parseValue(str, geo.z, '+');

	// Parse rotation
	str = parseValue(str, geo.xrot, '*');
	str = parseValue(str, geo.yrot, '*');
	str = parseValue(str, geo.zrot, '*');
    }

    void buildClock(Node_ptr root) {
	MaterialAttributes material;

	Geometry geo(10, 10, 1);
	std::string geo_str = _opt.get("geometry");
	if (geo_str != "") 
	    parseGeometry(geo_str, geo);

	// Create some primitives to use
	Primitive_var cylinder = _prim_kit->createCylinder(32);
	Primitive_var box = _prim_kit->createCuboid();
	
	// Create the clock face
	Appearance_var app1 = _geo_kit->createAppearance();
	Solid::Primitive_var face = _solid_kit->createPrimitive(cylinder);
	Solid::Appearance_var a_face = _solid_kit->createAppearance(app1);
	Solid::Transform_var t_face = _solid_kit->createTransform(a_face);


	t_face->scale(Nobelxx::Vertex3D(10.0, 1.0, 10.0));
	t_face->rotate(90, Nobelxx::xaxis);

	material.diffuse = Nobelxx::Color(0.3, 0.41, 0.51);
	app1->material(material);
	a_face->body(face);

	// Create the clock hands
	Appearance_var app2 = _geo_kit->createAppearance();
	Solid::Primitive_var hand = _solid_kit->createPrimitive(box);
	Solid::Appearance_var a_hand = _solid_kit->createAppearance(app2);
	Solid::Transform_var m_hand = _solid_kit->createTransform(a_hand);
	Solid::Transform_var h_hand = _solid_kit->createTransform(a_hand);
	_minutes = _solid_kit->createTransform(m_hand);
	_hours = _solid_kit->createTransform(h_hand);
	
	m_hand->scale(Nobelxx::Vertex3D(0.5, 8, 0.1));
	m_hand->translate(Nobelxx::Vertex3D(0, 3, 1.3));

	h_hand->scale(Nobelxx::Vertex3D(0.75, 4, 0.1));
	h_hand->translate(Nobelxx::Vertex3D(0, 1, 1.3));

	material.diffuse = Nobelxx::Color(0.8, 0.8, 0.8);
	app2->material(material);
	a_hand->body(hand);

	// Create an appropriate transform
	Solid::Transform_var tick = _solid_kit->createTransform(a_hand);
	tick->scale(Nobelxx::Vertex3D(0.4, 1, 0.2));
	tick->translate(Nobelxx::Vertex3D(0, 9, 1.1));

	// Now, combine it with the face
	Solid::Binary_var clock = _solid_kit->createUnion(tick, t_face);

	// Create the rest of the face
	for (int i = 1; i < 12; i++) {
	    
	    // Create an appropriate transform
	    tick = _solid_kit->createTransform(hand);

	    // Set the scale correctly
	    tick->scale(Nobelxx::Vertex3D(0.2, 0.5, 0.2));
	    tick->translate(Nobelxx::Vertex3D(0, 9, 1.1));
	    tick->rotate(360.0f * i / 12.0f, Nobelxx::zaxis);
	    
	    // Now, combine it with the face
	    clock = _solid_kit->createUnion(tick, clock);
	}

	// Combine the hands and the face
	Solid::Binary_var hands = _solid_kit->createUnion(_minutes, _hours);
	clock = _solid_kit->createUnion(hands, clock);
	
	// Now we create a transform to scale the clock accordingly
	Solid::Transform_var final = _solid_kit->createTransform(clock);
	final->scale(Nobelxx::Vertex3D(geo.width / 10,
				       geo.height / 10,
				       geo.depth / 10));
	final->rotate(geo.yrot, Nobelxx::yaxis);
	final->rotate(geo.xrot, Nobelxx::xaxis);
	final->rotate(geo.zrot, Nobelxx::zaxis);
	final->translate(Nobelxx::Vertex3D(geo.x, geo.y, geo.z));
	
	// Create solid container to insert into the scene graph
	_node = _solid_kit->createContainer();

	// Set CSG trees into the nodes
	_node->setTree(final);

	// Finally, insert them into the scene graph
	root->insert(_node);
    }
    
    void setHands(int hours, int minutes, int seconds) {
	
	// Compute values for the hours and minutes
	int min_in_sec = minutes * 60 + seconds;
	int hour_in_sec = hours * 60 * 60 + minutes * 60 + seconds;
	
	// Compute angles for the clock hands
	float min_angle = 360.0f * min_in_sec / (60.0f * 60.0f);
	float hour_angle = 360.0 * hour_in_sec / (12.0 * 60.0 * 60.0f);
	
	// Set transforms
	_minutes->identity();
	_hours->identity();
	_minutes->rotate(-min_angle, Nobelxx::zaxis);
	_hours->rotate(-hour_angle, Nobelxx::zaxis);
	
	// Re-evaluate the solid node
	_node->evaluate();
    }
    
    Nobelxx::Client _client;
    GetOpt _opt;
    SolidKit_var _solid_kit;
    PrimitiveKit_var _prim_kit;
    GeometryKit_var _geo_kit;
    Solid::Transform_var _minutes, _hours;
    SolidContainer_var _node;
};

// -- Code Segment

int main(int argc, char *argv[]) {
    
    try {

	// Create a command-line parser
	GetOpt opt("3Dwm Clock", VERSION, "[options]");
	
	// Add some command-line options
	opt.add('g', "geometry", GetOpt::mandatory,
		"application geometry (WxHxD+X+Y+Z*Xr*Yr*Zr)");

	// Parse the command-line
	if (opt.parse(argc, argv) == false)
	    return EXIT_FAILURE;

	// Initialize application
	Clock clock(argc, argv, opt);
	
	// Run it! (Never terminates)
	clock.run();
    }
    catch (...) {
	return EXIT_FAILURE;
    }
    
    return EXIT_SUCCESS;
}
