/* ------------------------------------------------------------------------
 * $Id: BoundingBox.hh,v 1.2 2002/03/20 15:45:58 elm Exp $
 *
 * This file is part of 3Dwm: The Three-Dimensional User Environment.
 *
 * 3Dwm: The Three-Dimensional User Environment:
 *	<http://www.3dwm.org>
 *
 * Chalmers Medialab
 * 	<http://www.medialab.chalmers.se>
 * 
 * ------------------------------------------------------------------------
 * File created 2001-12-05 by Niklas Elmqvist.
 *
 * Copyright (c) 2001 Niklas Elmqvist <elm@3dwm.org>.
 * ------------------------------------------------------------------------
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * ------------------------------------------------------------------------
 */

#ifndef _BoundingBox_hh_
#define _BoundingBox_hh_

// -- System Includes
#include <vector>

// -- Celsius Includes
#include "Matrix3D.hh"
#include "Vector3D.hh"
#include "Adapters.hh"

// -- Class Declarations

class BoundingBox {
public:
    
    /**
     * Constructor.
     **/ 
    BoundingBox();

    /**
     * Constructor.
     *
     * @param center center point of box.
     * @param extents extents of box.
     * @param transform transformation matrix of box.
     **/
    BoundingBox(const Vector3D &center, const Vector3D &extents,
		const Matrix3D &transform);
    
    /**
     * Destructor.
     **/ 
    virtual ~BoundingBox();
    
    template <class E>
    bool isInside(const VertexAdapter<E> &p) const;
    
    template <class S, class E>
    void update(const IndexedAdapter<S, E> &vs);
     
    static BoundingBox merge(const BoundingBox &box0, const BoundingBox &box1);

    const Vector3D &getCenter() const { return _center; }
    const Vector3D &getExtents() const { return _extents; }
    const Matrix3D &getTransform() const { return _transform; }

    void premultiplyTransform(const Matrix3D &m);
    void postmultiplyTransform(const Matrix3D &m);

    std::vector<Vector3D> getVertexList() const;

    bool operator == (const BoundingBox &b) const;
    
private:
    Vector3D _center, _extents;
    Matrix3D _transform;
};

template <class S, class E>
inline void BoundingBox::update(const IndexedAdapter<S, E> &vs)
{
    VertexAdapter<E> vertex(vs[0]);
    Vector3D vmin(vertex.x(), vertex.y(), vertex.z()), vmax = vmin;

    // Find the minimum and vmaximum vertices
    for (size_t i = 1; i < vs.length(); i++) {
	
	VertexAdapter<E> vertex(vs[i]);

	if (vertex.x() < vmin.x()) vmin.x() = vertex.x();
	if (vertex.y() < vmin.y()) vmin.y() = vertex.y();
	if (vertex.z() < vmin.z()) vmin.z() = vertex.z();

	if (vertex.x() > vmax.x()) vmax.x() = vertex.x();
	if (vertex.y() > vmax.y()) vmax.y() = vertex.y();
	if (vertex.z() > vmax.z()) vmax.z() = vertex.z();
    }
    
    // Find extents of the box
    _extents = (vmax - vmin) / 2;

    // Find the center point
    _center = _extents + vmin;
    
    // Now set the orthonormal axises (X, Y and Z), i.e the identity
    // matrix
    _transform.identity();
}

template <class E>
inline bool BoundingBox::isInside(const VertexAdapter<E> &p) const
{
    // Compute normalized vector
    Vector3D norm(p.x()-_center.x(), p.y()-_center.y(), p.z()-_center.z());
    
    // Multiply using the transform
    Vector3D res = _transform * norm;
    
    return std::abs(res.x()) <= _extents.x() &&
	std::abs(res.y()) <= _extents.y() &&
	std::abs(res.z()) <= _extents.z();
}

inline bool BoundingBox::operator == (const BoundingBox &b) const
{
    // Compare the spatial members
    return b._center == _center
	&& b._extents == _extents
	&& b._transform == _transform;
}

#endif /* BoundingBox.hh */
