/* ------------------------------------------------------------------------
 * $Id: GetOpt.hh,v 1.2 2001/11/17 22:20:01 elm Exp $
 *
 * This file is part of 3Dwm: The Three-Dimensional User Environment.
 *
 * 3Dwm: The Three-Dimensional User Environment:
 *	<http://www.3dwm.org>
 *
 * Chalmers Medialab
 * 	<http://www.medialab.chalmers.se>
 * 
 * ------------------------------------------------------------------------
 * File created 2001-11-12 by Niklas Elmqvist.
 *
 * Copyright (c) 2001 Niklas Elmqvist <elm@3dwm.org>.
 * ------------------------------------------------------------------------
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * ------------------------------------------------------------------------
 */

#ifndef _GetOpt_hh_
#define _GetOpt_hh_

// -- System Includes
#include <string>
#include <vector>

// -- Celsius Includes
#include "Exception.hh"

// -- Class Declarations

/**
 * Command-line options parsing class.
 **/
class GetOpt {
public:

    /// Option argument type
    typedef enum { optional, mandatory, no_value } opt_flags;
    
    /**
     * Constructor.
     *
     * @param name program name.
     * @param version program version.
     **/
    GetOpt(const std::string &name = "",
	   const std::string &version = "",
	   const std::string &help_msg = "[options]");

    /**
     * Emit a usage message (including help texts).
     *
     * @param arg0 invokation name.
     **/ 
    void usage(const std::string &arg0);

    /**
     * Emit a version message.
     **/ 
    void version();

    /**
     * Parse the command-line and assign values to options.
     * 
     * @param argc number of arguments in array.
     * @param argv array of arguments.
     * @return true if successful (no parse error), false otherwise.
     **/ 
    bool parse(int argc, char **argv);
    
    /**
     * Register a command-line option.
     *
     * @param short_key short command-line option key ("-k").
     * @param long_key long command-line option key ("--key").
     * @param flags option flags.
     * @param desc description (for help output).
     **/
    void add(char short_key, const std::string &long_key,
	     opt_flags flags, const std::string &desc = "no description");
    
    /**
     * Retrieve the value associated with a specific option. If the
     * option was not present on the command line, the empty string
     * will be returned. If the option was present, but no value was
     * specified, "true" (the string) will be returned.
     *
     * @param short_key short character to look up.
     * @return value associated with option (if any).
     **/ 
    const std::string &get(char short_key) const;

    /**
     * Retrieve the value associated with a specific option. If the
     * option was not present on the command line, the empty string
     * will be returned. If the option was present, but no value was
     * specified, "true" (the string) will be returned.
     *
     * @param long_key long argument to look up.
     * @return value associated with option (if any).
     **/ 
    const std::string &get(const std::string &long_key) const;
    
    /**
     * Retrieve the number of arguments
     *
     * @return number of arguments parsed.
     **/ 
    int getArgumentCount() const {
	return _arguments.size(); 
    }

    /**
     * Retrieve an argument using an index.
     **/
    const std::string getArgument(int i) const {
	if (i >= 0 && i < getArgumentCount())
	    return _arguments[i];
	else
	    throw Exception("index out of range!");
    }
    
private:

    struct option_t {
	char short_key;
	std::string long_key;
	opt_flags flags;
	std::string desc;
	std::string value;
	
	option_t(char sk, const std::string &lk, opt_flags f,
		 const std::string &d)
	    : short_key(sk), long_key(lk), flags(f), desc(d) { }
	
	bool operator == (const std::string &arg) const {
	    return arg == long_key;
	}
	bool operator == (char c) const {
	    return c == short_key;
	}
    };
    
    std::string _name;
    std::string _version;
    std::string _help_msg;
    std::vector<option_t> _options;
    std::vector<std::string> _arguments;
};


#endif /* GetOpt.hh */
