/* ------------------------------------------------------------------------ 
 * $Id: BoundingBox.cc,v 1.2 2002/03/20 15:45:59 elm Exp $
 *
 * This file is part of 3Dwm: The Three-Dimensional User Environment.
 *
 * 3Dwm: The Three-Dimensional User Environment:
 *	<http://www.3dwm.org>
 *
 * Chalmers Medialab
 * 	<http://www.medialab.chalmers.se>
 * 
 * ------------------------------------------------------------------------
 * File created 2001-12-05 by Niklas Elmqvist.
 *
 * Copyright (c) 2001 Niklas Elmqvist <elm@3dwm.org>.
 * ------------------------------------------------------------------------
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * ------------------------------------------------------------------------
 */

// -- System Includes
#include <algorithm>
#include <cmath>

// -- Celsius Includes
#include "Quaternion.hh"
#include "BoundingBox.hh"

// -- Code Segment

BoundingBox::BoundingBox() 
{
    // Orthonormal matrix of unit axis vectors
    _transform.identity();
}

BoundingBox::BoundingBox(const Vector3D &center, const Vector3D &extents,
			 const Matrix3D &transform)
    : _center(center), _extents(extents), _transform(transform)
{
    // empty
}

BoundingBox::~BoundingBox()
{
    // empty
}

BoundingBox BoundingBox::merge(const BoundingBox &box0,
			       const BoundingBox &box1)
{
    BoundingBox box;
    
    // Compute center
    box._center = (box0._center + box1._center) / 2;
    
    // Compute axes using quaternions
    Quaternion q0, q1;
    q0.fromMatrix(box0._transform);
    q1.fromMatrix(box1._transform);
    Quaternion q = q0 + q1;
    q.normalize();
    box._transform = q.toMatrix();

    // Compute new extents
    std::vector<Vector3D> vertexList[2];
    vertexList[0] = box0.getVertexList();
    vertexList[1] = box1.getVertexList();

    // Process each box
    for (int b = 0; b < 2; b++) {
	for (std::vector<Vector3D>::iterator i = vertexList[b].begin();
	     i != vertexList[b].end(); i++) {
	    
	    // Compute delta from new center
	    Vector3D delta = *i - box._center;
	    
	    // Project the point onto the new box
	    for (int j = 0; j < 3; j++) {
		
		// Retrieve pertinent axis from rotation transform
		Vector3D axis(box._transform(0, j),
			      box._transform(1, j),
			      box._transform(2, j));
		
		// Project the point on the box
		float abs_proj = Math::abs(Vector3D::dot(axis, delta));
		
		// If we are outside of the extents, grow them
		if (abs_proj > box._extents(j))
		    box._extents(j) = abs_proj;
	    }
	}
    }

    // Return with resulting box
    return box;
}

void BoundingBox::premultiplyTransform(const Matrix3D &m)
{
    _transform = m * _transform;
}

void BoundingBox::postmultiplyTransform(const Matrix3D &m)
{
    _transform = _transform * m;
}

std::vector<Vector3D> BoundingBox::getVertexList() const
{
    float vertices[8][3] = {
	{  _extents.x(),  _extents.y(),  _extents.z() },
	{  _extents.x(),  _extents.y(), -_extents.z() },
	{  _extents.x(), -_extents.y(),  _extents.z() },
	{  _extents.x(), -_extents.y(), -_extents.z() },
	{ -_extents.x(),  _extents.y(),  _extents.z() },
	{ -_extents.x(),  _extents.y(), -_extents.z() },
	{ -_extents.x(), -_extents.y(),  _extents.z() },
	{ -_extents.x(), -_extents.y(), -_extents.z() }
    };
    std::vector<Vector3D> vertexList;
    
    // Compute the box vertices
    for (int i = 0; i < 8; i++) {
	Vector3D vertex(vertices[i][0], vertices[i][1], vertices[i][2]);

	// Vertex is V = C + R * E (where E is box extents)
	vertex = _transform * vertex;
	vertex += _center;

	// Add to the list
	vertexList.push_back(vertex);
    }
    return vertexList;
}
