/* ------------------------------------------------------------------------
 * $Id: GetOpt.cc,v 1.1 2001/11/13 15:18:07 elm Exp $
 *
 * This file is part of 3Dwm: The Three-Dimensional User Environment.
 *
 * 3Dwm: The Three-Dimensional User Environment:
 *	<http://www.3dwm.org>
 *
 * Chalmers Medialab
 * 	<http://www.medialab.chalmers.se>
 * 
 * ------------------------------------------------------------------------
 * File created 2001-11-12 by Niklas Elmqvist.
 *
 * Copyright (c) 2001 Niklas Elmqvist <elm@3dwm.org>.
 * ------------------------------------------------------------------------
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * ------------------------------------------------------------------------
 */

// -- System Includes
#include <algorithm>

// -- Local Includes
#include "Exception.hh"
#include "GetOpt.hh"

// -- Code Segment

GetOpt::GetOpt(const std::string &name,
	       const std::string &version,
	       const std::string &help_msg)
    : _name(name),
      _version(version),
      _help_msg(help_msg)
{
    // empty
}

bool GetOpt::parse(int argc, char **argv)
{
    int i = 1; 

    // Process all command line arguments
    while (i < argc) {

	// Retrieve the current argument
	std::string arg = argv[i];

	// Intercept help requests
	if (arg == "-h" || arg == "--help" || arg == "-?") {
	    usage(argv[0]);
	    return false;
	}
	// Intercept version requests as well
	else if (arg == "-v" || arg == "--version") {
	    version();
	    return false;
	}
	// Is this a command-line option? 
	else if (arg[0] == '-') {
	    
	    std::string value;
	    
	    // Remove the value from the argument (if there is one)
	    std::string::size_type pos = arg.find('=');
	    if (pos != std::string::npos) {
		value = arg.substr(pos + 1);
		arg.erase(pos);
	    }

	    // Find the option we are looking for (if it is there)
	    std::vector<option_t>::iterator opt = _options.end();

	    // Is it a long or a short option?
	    if (arg.length() > 1) {
		
		if (arg[1] == '-')
		    opt = std::find(_options.begin(), _options.end(),
				    arg.substr(2));
		else
		    opt = std::find(_options.begin(), _options.end(), arg[1]);
	    }
	    
	    // If we didn't find it, we have a parse error
	    if (opt == _options.end()) {
		std::cerr << argv[0] << ": invalid option -- " 
			  << arg << std::endl;
		std::cerr << "Try `" << argv[0] 
			  << " --help' for more information."
			  << std::endl;
		return false;
	    }
	    
	    // Isolate the value (after a '=' sign) if necessary
	    if (opt->flags != no_value) {

		// Use the given value if it exists
		if (value != "") {
		    opt->value = value;
		}
		else {
		    // Then, use the following argument as value
		    if ((i + 1) < argc) // && argv[i + 1][0] != '-')
			opt->value = argv[++i];
		    else {

			// Mandatory options requires values
			if (opt->flags == mandatory) {
			    std::cerr << argv[0] 
				      << ": expected value for argument "
				      << arg << std::endl;
			    std::cerr << "Try `" << argv[0] 
				      << " --help' for more information."
				      << std::endl;
			    return false;
			}
			// Optional options do not require values
			else {
			    opt->value = "true";
			}
		    }
		}
	    }
	    else {
		opt->value = "true";
	    }
	}
	// Okay, then it is a naked argument
	else {
	    _arguments.push_back(arg);
	}

	// Move forward to the next argument
	i++;
    }

    return true;
}

void GetOpt::usage(const std::string &binary)
{
    std::cerr << "Usage: " << binary << _help_msg << std::endl;
    for (std::vector<option_t>::iterator i = _options.begin(); 
	 i != _options.end(); i++) {
	std::cerr << "  ";
	if (i->short_key != '\0')
	    std::cerr << "-" << i->short_key;
	std::cerr << ", --" << i->long_key << "\t" << i->desc << std::endl;
    }
}

void GetOpt::version()
{
    std::cerr << _name << " " << _version << std::endl;
}

void GetOpt::add(char short_key, const std::string &long_key,
		 GetOpt::opt_flags flags, const std::string &desc)
{
    // Add the option to the option table
    _options.push_back(option_t(short_key, long_key, flags, desc));
}

const std::string &GetOpt::get(char short_key) const
{
    std::vector<option_t>::const_iterator opt 
	= std::find(_options.begin(), _options.end(), short_key);
    if (opt == _options.end()) 
	throw Exception("Option not found!");
    return opt->value;
}

const std::string &GetOpt::get(const std::string &long_key) const
{
   std::vector<option_t>::const_iterator opt 
       = std::find(_options.begin(), _options.end(), long_key);
    if (opt == _options.end()) 
	throw Exception("Option not found!");
    return opt->value;
}
