/* ------------------------------------------------------------------------ 
 * $Id: VolumeImpl.cc,v 1.7 2002/03/22 10:32:25 elm Exp $
 *
 * This file is part of 3Dwm: The Three-Dimensional User Environment.
 *
 * 3Dwm: The Three-Dimensional User Environment:
 *	<http://www.3dwm.org>
 *
 * Chalmers Medialab
 * 	<http://www.medialab.chalmers.se>
 * 
 * ------------------------------------------------------------------------
 * File created 2001-05-29 by Niklas Elmqvist.
 *
 * Copyright (c) 2001 Niklas Elmqvist <elm@3dwm.org>.
 * Copyright (c) 2001 Steve Houston <steve@3dwm.org>.
 * ------------------------------------------------------------------------
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * ------------------------------------------------------------------------
 */

// -- System Includes
#include <algorithm>
#include <cmath>

// -- 3Dwm Includes
#include "Celsius/Mutex.hh"
#include "Celsius/Guard.hh"
#include "Celsius/Logger.hh"
#include "Polhem/VolumeImpl.hh"

using namespace Nobel;

// -- Adapters

template <>
class IndexedAdapter<Vertex3DSeq, Vertex3D> {
public:
    IndexedAdapter(const Vertex3DSeq &impl) : _impl(impl) { }
    const Vertex3D & operator [] (int n) const { return _impl[n]; }
    size_t length() const { return _impl.length(); }
private:
    const Vertex3DSeq &_impl;
};

template <>
class VertexAdapter<Vertex3D> {
public:
    VertexAdapter(const Vertex3D &v) : _v(v) { }
    float x() const { return _v.x; }
    float y() const { return _v.y; }
    float z() const { return _v.z; }
private:
    const Vertex3D &_v;
};

// -- Code Segment

VolumeImpl::VolumeImpl()
{
    // empty
}

VolumeImpl::VolumeImpl(const VolumeImpl &v)
{
    *this = v;
}

VolumeImpl::~VolumeImpl()
{
    // empty
}
/*
Vertex3D VolumeImpl::getCenter()
{
    Vector3D c = _box.getCenter();
    Vertex3D center = { c.x(), c.y(), c.z() };
    return center;
}

Vertex3D VolumeImpl::getExtents()
{  
    Vector3D e = _box.getExtents();
    Vertex3D extents = { e.x(), e.y(), e.z() };
    return extents;
}

void VolumeImpl::getTransformationMatrix(Matrix m)
{
    _box.getTransform().store(m);
}
*/

Nobel::Volume::Box VolumeImpl::getBox()
{
    Nobel::Volume::Box b;
    
    // Store values into box
    b.center.x = _box.getCenter().x();
    b.center.y = _box.getCenter().y();
    b.center.z = _box.getCenter().z();

    b.extents.x = _box.getExtents().x();
    b.extents.y = _box.getExtents().y();
    b.extents.z = _box.getExtents().z();

    _box.getTransform().store(b.transform);

    return b;
}

CORBA::Boolean VolumeImpl::isInside(const Vertex3D &v)
{
    // Delegate to bounding box
    return _box.isInside(VertexAdapter<Vertex3D>(v));
}

void VolumeImpl::update(const Vertex3DSeq &vertices)
{
    // Delegate to bounding box
    _box.update(IndexedAdapter<Vertex3DSeq, Vertex3D>(vertices));
}

void VolumeImpl::mergeWithBox(const Nobel::Volume::Box &b)
{
    /*
    // Some volumes may be empty; check for this
    if (CORBA::is_nil(v))
	return; 
    */

    // Create a bounding box out of the target volume
    BoundingBox target(Vector3D(b.center.x, b.center.y, b.center.z),
		       Vector3D(b.extents.x, b.extents.y, b.extents.z),
		       b.transform);
    
    // Delegate to bounding boxes to perform the merge
    _box = BoundingBox::merge(_box, target);
}

void VolumeImpl::render(Nobel::Renderer_ptr r)
{
    Nobel::LineMesh mesh;
    
    // Retrieve vertices and build point list
    std::vector<Vector3D> vecList = _box.getVertexList();
    mesh.vertexList.length(vecList.size());
    for (size_t i = 0; i < vecList.size(); i++) {
	Vertex3D point = { vecList[i].x(), vecList[i].y(), vecList[i].z() };
	mesh.vertexList[i] = point;
    }

    // Now we want to build the line index list
    mesh.vertexIndexList.length(12 * 2);
    int lineIndexList[12][2] = {
	{ 0, 1 }, { 1, 3 }, { 3, 2 }, { 2, 0 },
	{ 4, 5 }, { 5, 7 }, { 7, 6 }, { 6, 4 },
	{ 0, 4 }, { 1, 5 }, { 2, 6 }, { 3, 7 }
    };
    for (size_t j = 0; j < 12; j++) {
	mesh.vertexIndexList[j * 2] = lineIndexList[j][0];
	mesh.vertexIndexList[j * 2 + 1] = lineIndexList[j][1];
    }

    // Render the line mesh
    r->renderLines(mesh);
}
